
/*
 * code.c -- main routines for PIC16C84 Assembler
 *
 * Copyright (C) 1995-6 James Cleverdon
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Address:
 *	James Cleverdon
 *	16465 S.W. Estuary Dr. # 101
 *	Beaverton, OR 97006-7954, USA
 *
 *	jamesc@sequent.com
 *
 * $Log:	code.c,v $
 * Revision 1.3  96/01/25  00:49:41  jamesc
 * Put the code_mem, etc into a structure in case
 * a compiler decides to move them around (like
 * Sequent's).  Improved listing format slightly.
 * 
 * Revision 1.2  96/01/03  11:46:45  jamesc
 * Added support for rudimentary listings.  Deleted
 * line_buf[] and added list_head/list_tail.  NODEs
 * point at their source code lines, which point to
 * the resulting memory value (if any) using
 * n_extra.sp.
 *
 * Revision 1.0  1995/12/19 01:28:34  jamesc
 * "CSE510_project"
 *
 */

#ident "$Id: code.c,v 1.3 96/01/25 00:49:41 jamesc Exp Locker: jamesc $"

#include <sys/types.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include "c84.h"
#include "c84_ext.h"
#include "c84_mem.h"
#include "x.tab.h"

#define DEBUG		/* extra checks + verbose */
/* #define FULL_DEBUG	/* very verbose */
#define TREE_DEBUG	/* debug parse trees */

#define Static		/* some source code debuggers don't like statics */

#ifdef NO_BZERO
#define bzero(adr, num) memset((void *)(adr), 0, (num))
#endif

#define ALLOC_NODES	32		/* # of nodes to alloc at a time */

#define NUM_HASH	32		/* # of symbol hash buckets	*/
#define SYM_HASH(name)	(&sym_hash[hash_val(name)])

/*
 * Token type defines.
 */

#define IS_BINOP(np)	((np)->n_flags & F_BIN_OP)
#define IS_CODE(np)	((np)->n_flags & F_OP_CD)
#define IS_CONST(np)	((np)->n_flags & F_IS_CONST)
#define IS_EXPR(np)	((np)->n_flags & (F_IS_VAL | F_BIN_OP | F_UNI_OP))
#define IS_OPER(np)	((np)->n_flags & (F_BIN_OP | F_UNI_OP))
#define IS_UNIOP(np)	((np)->n_flags & F_UNI_OP)
#define IS_VALUE(np)	((np)->n_flags & F_IS_VAL)

/*
 * token names
 */

Static char	*names[] = {
/*00*/ "EOF", "ADDLW", "ADDWF", "ANDLW", "ANDWF", "BCF", "BSF", "BTFSC",
/*08*/ "BTFSS", "CALL", "\\n", "CLRF", "CLRW", "CLRWDT", "COMF", "DECF",
/*10*/ "DECFSZ", "GOTO", "INCF", "INCFSZ", "IORLW", "IORWF", "MOVF", "MOVLW",
/*18*/ "MOVWF", "NOP", "OPTION", "RETFIE", "RETLW", "RETURN", "RLF", "RRF",
/*20*/ "SLEEP", "!", "\"", "ID", "$", "%", "&", "'",
/*28*/ "(", ")", "*", "+", ",", "-", ".", "/",
/*30*/ "0", "1", "2", "3", "4", "5", "6", "7",
/*38*/ "8", "9", ":", ";", "<", "=", ">", "_",
/*40*/ "@", "SUBLW", "SUBWF", "SWAPF", "TRIS", "XORLW", "XORWF", "BIT",
/*48*/ "DBYTE", "DORG", "DWORD", "EBYTE", "END", "EQU", "EORG", "EWORD",
/*50*/ "FILL", "FUSE", "LINE", "ORG", "SET", "B_CONST", "CONST", "LABEL",
/*58*/ "UNDEF_LABEL", "STRING", "||", "&&", "==", "!=", "^", "_",
/*60*/ "`", ">=", "<=", "<<", ">>", "MAX", "MIN", "HIGH",
/*68*/ "LOW", "U-", "j", "k", "l", "m", "n", "o",
/*70*/ "p", "q", "r", "s", "F_B", "F_D", "LINE", "w",
/*78*/ "x", "y", "z", "{", "|", "}", "~", "DEL",
};

/*
 * PIC 16C84 Register and Bit names -- initial values for the symbol table
 */

typedef struct sym_init_s {
	char	*si_name;
	ushort	si_value;
	uchar	si_tok;
} SYM_INIT;

#define I_BIT(name, reg, bit)	{ name, BIT_VAL(reg, bit), T_B_CONST }
#define I_EQU(name, value)	{ name, value, T_CONST }

Static SYM_INIT sym_init[] = {
  /* Page 0 Registers */
I_EQU("IND0",	0x00),	/* Indirects via FSR into data memory	*/
I_EQU("RTCC",	0x01),	/* Real Time Clock Counter		*/
I_EQU("PCL",	0x02),	/* PC's Low 8 bits			*/
I_EQU("STATUS",	0x03),	/* Status register			*/
I_EQU("FSR",	0x04),	/* Data address for indirect via IND0	*/
I_EQU("PORT_A",	0x05),	/* Port A data bits			*/
I_EQU("PORT_B",	0x06),	/* Port B data bits			*/
/* 07 is Port C in some PIC parts, but not 16C84		*/
I_EQU("EEDATA",	0x08),	/* EEROM data port			*/
I_EQU("EEADR",	0x09),	/* EEROM address register		*/
I_EQU("PCLATH",	0x0A),	/* Holding reg for PC's High byte (see manual)	*/
I_EQU("INTCON",	0x0B),	/* Interrupt Control register		*/

  /* Page 1 Registers	*/

/* IND0	 at	0x80	is mapped to page 0			*/
I_EQU("OPTION",	0x81),	/* Option select register		*/
/* PCL	 at	0x82	is mapped to page 0			*/
/* STATUS at	0x83	is mapped to page 0			*/
/* FSR	 at	0x84	is mapped to page 0			*/
I_EQU("TRIS_A",	0x85),	/* Port A data direction bits		*/
I_EQU("TRIS_B",	0x86),	/* Port B data direction bits		*/
/* 87 is unimplemented in 16C84	*/
I_EQU("EECON1",	0x88),	/* EEROM control register		*/
I_EQU("EECON2",	0x89),	/* not a physical register		*/
/* PCLATH at	0x8A	is mapped to page 0			*/
/* INTCON at	0x8B	is mapped to page 0			*/

/* STATUS bits	*/
I_EQU("CARRY",	0),		/* Carry/Borrow* bit (see manual)	*/
I_BIT("C_bit",	0x03,0),	/* Carry/Borrow* bit (see manual)	*/
I_EQU("DC",	1),		/* Digit Carry/Borrow* bit (see manual)	*/
I_BIT("DC_bit",	0x03,1),	/* Digit Carry/Borrow* bit (see manual)	*/
I_EQU("Z",	2),
I_BIT("Z_bit",	0x03,2),	/* Zero bit				*/
I_BIT("PD_bit",	0x03,3),	/* Power Down status bit		*/
I_EQU("PD",	3),
I_BIT("TO_bit",	0x03,4),	/* Time Out status bit			*/
I_EQU("TO",	4),
I_EQU("RP0",	5),		/* RP1,0: Register Page for direct addr	*/
I_BIT("RP0_bit",0x03,5),	/*   0,0: Pg 0 (00-7F); 0,1: Pg 1 (80-8F) */
I_EQU("RP1",	6),
I_BIT("RP1_bit",0x03,6),
I_EQU("IRP0",	7),		/* Indirect Reg Page select (not used yet) */
I_BIT("IRP0_bit", 0x03,7),

/* PORT_A bits	*/
I_BIT("RA0",	0x05,0),	/* PORT A, bit 0			*/
I_BIT("RA1",	0x05,1),	/* PORT A, bit 1			*/
I_BIT("RA2",	0x05,2),	/* PORT A, bit 2			*/
I_BIT("RA3",	0x05,3),	/* PORT A, bit 3			*/
I_BIT("RA4",	0x05,4),	/* PORT A, bit 4 (RTCC input)		*/
I_BIT("RT",	0x05,4),	/* PORT A, bit 4 (RTCC input)		*/

/* PORT_B bits	*/
I_BIT("RB0",	0x06,0),	/* PORT B, bit 0 (INT input)		*/
I_BIT("RB1",	0x06,1),	/* PORT B, bit 1			*/
I_BIT("RB2",	0x06,2),	/* PORT B, bit 2			*/
I_BIT("RB3",	0x06,3),	/* PORT B, bit 3			*/
I_BIT("RB4",	0x06,4),	/* PORT B, bit 4 (RB4-7 intr-on-change)	*/
I_BIT("RB5",	0x06,5),	/* PORT B, bit 5			*/
I_BIT("RB6",	0x06,6),	/* PORT B, bit 6			*/
I_BIT("RB7",	0x06,7),	/* PORT B, bit 7			*/

/* INTCON bits	*/
I_BIT("RBIF",	0x0B,0),	/* INTCON, bit 0 (Change on RB pend Flg)*/
I_BIT("INTF",	0x0B,1),	/* INTCON, bit 1 (INT pending Flag)	*/
I_BIT("RTIF",	0x0B,2),	/* INTCON, bit 2 (RTCC Intr pending Flg)*/
I_BIT("RBIE",	0x0B,3),	/* INTCON, bit 3 (Change on RB4-7 IntEn)*/
I_BIT("INTE",	0x0B,4),	/* INTCON, bit 4 (INT Intr Enable)	*/
I_BIT("RTIE",	0x0B,5),	/* INTCON, bit 5 (RTCC overflow Intr En)*/
I_BIT("EEIE",	0x0B,6),	/* INTCON, bit 6 (EEwrite Intr Enable)	*/
I_BIT("GIE",	0x0B,7),	/* INTCON, bit 7 (General Intr Enable)	*/

/* OPTION bits	*/
I_BIT("PS0",	0x81,0),	/* OPTION, bit 0 (Prescaler val bit 0)	*/
I_BIT("PS1",	0x81,1),	/* OPTION, bit 1 (Prescaler val bit 1)	*/
I_BIT("PS2",	0x81,2),	/* OPTION, bit 2 (Prescaler val bit 2)	*/
I_BIT("PSA",	0x81,3),	/* OPTION, bit 3 (Prescaler Assign bit)	*/
I_BIT("RTE",	0x81,4),	/* OPTION, bit 4 (RTCC signal edge)	*/
I_BIT("RTS",	0x81,5),	/* OPTION, bit 5 (RTCC signal source)	*/
I_BIT("INTEDG",	0x81,6),	/* OPTION, bit 6 (INT edge select)	*/
I_BIT("RBPU",	0x81,7),	/* OPTION, bit 7 (Port B pull-up Enable)*/

/* EECON1 bits	*/
I_BIT("RD",	0x88,0),	/* EECON1, bit 0 (Read ctl: 1=read)	*/
I_BIT("WR",	0x88,1),	/* EECON1, bit 1 (Write ctl: 1=write)	*/
I_BIT("WREN",	0x88,2),	/* EECON1, bit 2 (Write enable: 1=write)*/
I_BIT("WRERR",	0x88,3),	/* EECON1, bit 3 (Write error: 1=fail)	*/
I_BIT("EEIF",	0x88,4),	/* EECON1, bit 4 (EEwrite done Intr Flg)*/

/* Helpful defines	*/

I_EQU("Same",	1),	/* destination codes	*/
I_EQU("F",	1),
I_EQU("W",	0),
I_EQU("LSB",	0),
I_EQU("MSB",	7),

I_EQU("TRUE",	1),
I_EQU("FALSE",	0),
I_EQU("YES",	1),
I_EQU("NO",	0),

/* end of "picreg.h"	*/
};

long		line_num = 1;
int		col;
char		*this_file;

NODE		*list_head = NNULL;
NODE		*list_tail = NNULL;

Static char	*hex_file;
Static char	*lst_file;

Static ulong	code_addr = CODE_ORIGIN;
Static ulong	data_addr = DATA_ORIGIN;
Static ulong	eedata_addr = EEDATA_ORIGIN;
Static ulong	id_addr = ID_ORIGIN;

	/* The order *_mem is important to sym_output, so put in a struct */

Static struct all_mem_s {
	short	am_code_mem[CODE_MAX + 1];
	short	am_eedata_mem[EEDATA_MAX + 1];
	short	am_id_mem[ID_MAX + 1];
	short	am_fuse_mem;
} all_mem;

	/* for convenience */

#define	code_mem	all_mem.am_code_mem
#define	eedata_mem	all_mem.am_eedata_mem
#define	id_mem		all_mem.am_id_mem
#define	fuse_mem	all_mem.am_fuse_mem

Static NODE	*free_head = NNULL;
Static NODE	*free_tail = NNULL;

Static SYMBOL	*sym_hash[NUM_HASH];

Static uchar	*str_ptr;
Static uchar	string[LINE_LEN];

extern int	yydebug;


/*
 * tok_name -- return a token's name string
 */

char *
tok_name(int tok)
{
	static char	str[4];

	if (tok < NEL(names)) {
		return (names[tok]);
	}
	(void) sprintf(str, "$%02X", tok & 0xFF);
	return (str);
}


/*
 * prtable_char -- nicely display control and other non-printable chars
 */

char *
prtable_char(register uint c)
{
	static char	str[16];

	if (!isascii(c) || (!isprint(c) && !iscntrl(c))) {
		(void) sprintf(str, "(0x%02X)", c);
	} else if (iscntrl(c)) {
		(void) sprintf(str, "'^%c' (0x%02X)", c + '@', c);
	} else {
		(void) sprintf(str, "'%c' (0x%02X)", c, c);
	}
	return(str);
}


/*
 * alloc -- error trapped malloc with bzero
 */

void *
alloc(int size, char *msg)
{
	void	*vp;

	if (vp = malloc(size)) {
		bzero(vp, size);
		return (vp);
	}
	err("alloc: can't malloc any more memory (%s)!\n", msg);
	exit(1);
	return ((void *)NULL);
}


/*
 * more_nodes -- alloc a batch more nodes when the free list is empty
 */

Static NODEP
more_nodes(void)
{
	register NODEP	new;
	register int	n;

	new = (NODE *) alloc(sizeof(NODE) * ALLOC_NODES, "more_nodes");
	free_tail = new + 1;
	for (n = ALLOC_NODES; --n > 0; new++) {
		new->n_next = free_head;
		free_head = new;
	}
	return (new);				/* return last node */
}


/*
 * new_node -- hand out new NODEs
 */

NODEP
new_node(int tok)
{
	register NODEP	new;

	new = free_head;
	if (new) {
		free_head = new->n_next;
		/*
		 * Don't need to update free_tail, since we always check
		 * free_head first.
		 */
	}
	else {
		new = more_nodes();
	}
	new->n_left = NNULL;
	new->n_tok = (uchar) tok;
	new->n_lineno = line_num;
	new->n_col = col;
	new->n_file = this_file;
	new->n_line = list_tail;
	return (new);
}


/*
 * new_tok -- get a new NODE with the given token, and update col
 */

int
new_tok(int tok)
{
	yylval = new_node(tok);
	col += yyleng;
	return (tok);
}


/*
 * new_code -- get a new NODE with the given token and opcode value
 */

int
new_code(int tok, int opcode)
{
	register NODEP	np;

	yylval = np = new_node(tok);
	np->n_flags |= F_OP_CD;
	np->n_opcode = opcode;
	col += yyleng;
	return (tok);
}


/*
 * free_node -- toss old nodes onto tail of free list
 */

Static void
free_node(NODE *np)
{
	if (np == NNULL)
		return;

	if (np->n_tok == T_STRING && np->n_val.str != NULL) {
		free(np->n_val.str);
	}
#ifdef DEBUG
	/*
	 * Clear node to catch dangling references (system dependent).
	 */
	(void) bzero(np, sizeof(*np));
#endif
	if (free_head == NNULL) {
		free_head = np;
	}
	else {
		free_tail->n_next = np;
	}
	free_tail = np;
	np->n_next = NNULL;
}


/*
 * hash_val -- compute caseless hash value
 */

Static uint
hash_val(char *namep)
{
	register uchar 	*a = (uchar *) namep;
	register uint 	va;
	register uint	val;

	val = 0;
	while ((va = *a++) != '\0') {
		if (islower(va)) {
			va = toupper(va);
		}
		val += va;
	}

	return (val & (NUM_HASH - 1));
}


/*
 * cmp_ncase -- like strcmp, but compare two strings caselessly
 */

Static int
cmp_ncase(const void *ap, const void *bp)
{
	register uchar 	*a = (uchar *) ap;
	register uchar 	*b = (uchar *) bp;
	register uchar 	va, vb;

	do {
		va = *a++;
		vb = *b++;
		if (islower(va)) {
			va = toupper(va);
		}
		if (islower(vb)) {
			vb = toupper(vb);
		}
		if (va != vb) {
			return ((int)va - (int)vb);	/* different */
		}
	} while (va != '\0');

	return (0);					/* matched */
}


/*
 * label_scan -- look up a symbol in the list; if not found, add it
 */

NODEP
label_scan(char *name)
{
	register NODEP	np;
	register SYMBOL *sp;
	register SYMBOL **hashp;
	register int	len;
	char		str[SYM_LEN];

	np = new_node(T_CONST);
	len = yyleng;
	col += len;

	if (name[len - 1] == ':') {
		--len;				/* drop trailing colons */
		if (len >= SYM_LEN) {
			len = SYM_LEN - 1;
		}
		name = strncpy(str, name, len);	/* make working copy	*/
		name[len] = '\0';
	}

	hashp = SYM_HASH(name);
	for (sp = *hashp; sp; sp = sp->s_next) {
		if (cmp_ncase(name, sp->s_name) == 0) {
			break;
		}
	}

	if (sp) {
		/* found it */
		np->n_sym = sp;
		sp->s_flags &= ~F_DONT_LIST;		/* has been used */
		if (sp->s_flags & F_UNDEF_LABEL) {
			np->n_tok = T_UNDEF_LABEL;
			np->n_flags = F_IS_VAL | F_UNDEF_LABEL;
			np->n_next = sp->s_val.node;
			sp->s_val.node = np;		/* onto undef list */
		}
		else {
			np->n_flags = F_IS_VAL | F_IS_CONST;
			np->n_val.l = sp->s_val.l;
			np->n_tok = sp->s_type;
			if (np->n_tok == T_B_CONST) {
				np->n_extra.l = np->n_val.l;
				np->n_val.l = 1 << (np->n_val.l >> 7);
			}
		}
		return (np);
	}

	sp = (SYMBOL *) alloc(sizeof(SYMBOL) + len, "symbol add");
	(void) strncpy(sp->s_name, name, len);
	sp->s_name[len] = '\0';
	sp->s_next = *hashp;				/* link on hash list */
	*hashp = sp;
	sp->s_flags = F_UNDEF_LABEL;
	sp->s_col = np->n_col;
	sp->s_lineno = np->n_lineno;
	sp->s_val.node = np;				/* onto undef list */
	np->n_tok = T_UNDEF_LABEL;
	np->n_flags = F_IS_VAL | F_UNDEF_LABEL;
	np->n_sym = sp;
	return (np);
}


/*
 * backslash -- process backslash escaped character literals
 *
 *	Returns a constant from 0 to 255 and an updated spp, or an error.
 *
 *	Assumes that the backslash char has already been matched.
 */

int
backslash(uchar **spp)
{
	int	c;
	uchar	*up;
	int	cnt;
	int	num;

	up = *spp;
	if ((c = *up++) == '\0') {
		nerr(NNULL, "Unexpected end-of-file in backslash processing!");
	}

	switch (c) {
	case '\\':
	case '\'':
	case '"':
		*spp = up;
		return (c);
	case 'a':			/* "attention" ASCII BEL */
		*spp = up;
		return ('\007');
	case 'b':
		*spp = up;
		return ('\b');
	case 'f':
		*spp = up;
		return ('\f');
	case 'n':
		*spp = up;
		return ('\n');
	case 'r':
		*spp = up;
		return ('\r');
	case 't':
		*spp = up;
		return ('\t');
	case '0':	case '1':	case '2':	case '3':
	case '4':	case '5':	case '6':	case '7':
		num = 0;
		cnt = 3;		/* max of 3 octal digits */
		do {
			num <<= 3;
			num += c - '0';
		} while ((c = *up++) >= '0' && c <= '7' && --cnt > 0);
		if (c > 0)
			--up;		/* give back last char	*/
		*spp = up;
		return (num);
	case 'x':
		num = 0;
		cnt = 2;		/* max of 2 hexadecimal digits */
		while ((c = *up++) > 0 && isxdigit(c) && --cnt >= 0) {
			num <<= 4;
			if (isdigit(c))
				num += c - '0';
			else
				num += toupper(c) - 'A' + 10;
		}
		if (c > 0)
			--up;		/* give back last char	*/
		*spp = up;
		return (num);
	case '\n':
		nerr(NNULL, "New-line in string constant!");
	}
	err("Unrecognised escaped char 0x%02X; will use it anyway.", c);
	*spp = up;
	return(c);
}


/*
 * esc_const -- convert str into a escaped character constant
 */

int
esc_const(char *str)
{
	NODEP	np;

	yylval = np = new_node(T_CONST);
	np->n_flags |= F_IS_VAL | F_IS_CONST;
	np->n_val.l = backslash((uchar **) &str);
	col += yyleng;

	return (T_CONST);
}


/*
 * str_proc -- add bytes from yytext to string, converting escaped chars
 */

Static void
str_proc(register uchar *srcp, register int len)
{
	register uchar	*destp;
	register int	c;
	uchar		*sp;

	destp = str_ptr;

	while (--len >= 0) {
		++col;
		c = *srcp++;
		if (c == '\\') {
			sp = srcp;
			c = backslash(&sp);
			col += (sp - srcp);
			len -= (sp - srcp);
			srcp = sp;
		}
		if (destp >= &string[sizeof(string) - 1]) {
			nerr(NNULL, "String constant too long!");
		}
		*destp++ = (uchar) c;
	}

	str_ptr = destp;
}


/*
 * begin_string_const -- start a string constant
 */

void
begin_string_const(void)
{
	str_ptr = string;
	++col;					/* skip starting '"' */
	str_proc((uchar *) yytext + 1, yyleng - 1);
}


/*
 * more_string_const -- add to a string constant
 */

void
more_string_const(void)
{
	str_proc((uchar *) yytext, yyleng);
}


/*
 * end_string_const -- return a NODE for a good string constant, else exit
 */

int
end_string_const(void)
{
	NODEP	np;
	int	n;

	str_proc((uchar *) yytext, yyleng);
	if (*--str_ptr != '"') {
		nerr(NNULL, "New-line found in string constant!");
	} else {
		*str_ptr = '\0';
	}
	yylval = np = new_node(T_STRING);
	np->n_flags = F_IS_VAL | F_IS_CONST;
	n = (str_ptr - string) + 1;
	np->n_extra.l = n - 1;				/* strlen */
	np->n_val.str = alloc(n, "String constant");
	(void) strcpy(np->n_val.str, (char *) string);
	return (T_STRING);
}


/*
 * char_const -- convert the char at str into a character constant
 */

int
char_const(char *str)
{
	NODEP	np;

	yylval = np = new_node(T_CONST);
	np->n_flags |= F_IS_VAL | F_IS_CONST;
	np->n_val.l = *str;
	col += yyleng;

	return (T_CONST);
}


/*
 * const_n -- convert str to a base N constant
 */

int
const_n(char *str, int base)
{
	NODEP	np;
	char	*p;

	yylval = np = new_node(T_CONST);
	np->n_flags |= F_IS_VAL | F_IS_CONST;
	col += yyleng;

	p = str;
	if ((np->n_val.l = strtol(str, &p, base)) == 0 && p == str) {
		nerr(np, "const_n: no number!");
	}

	return (T_CONST);
}


/*
 * bin_op -- binary operation:  link lp and rp to np; return np
 */

NODEP
bin_op(NODEP lp, register NODEP np, NODEP rp)
{
	if (np == NNULL) {
		nerr(NNULL, "NULL binary operator?!?");
	}
	if (lp == NNULL) {
		nerr(np, "Binary operator missing left value!");
	}
	if (rp == NNULL) {
		nerr(np, "Binary operator missing right value!");
	}
	np->n_right = rp;
	rp->n_parent = np;
	np->n_left = lp;
	lp->n_parent = np;
	np->n_flags |= F_BIN_OP;
	return (np);
}


/*
 * un_op -- unary operation:  link arg to np; return np
 */

NODEP
un_op(register NODEP np, NODEP arg)
{
	if (np == NNULL) {
		nerr(NNULL, "NULL unary operator?!?");
	}
	if (arg == NNULL) {
		nerr(np, "Unary operator missing argument!");
	}
	np->n_right = arg;
	arg->n_parent = np;
	np->n_flags |= F_UNI_OP;
	return (np);
}


/*
 * comma_list -- build comma separated list, replicating opcdp
 */

NODEP
comma_list(NODEP opcdp, NODEP commap, NODEP valp)
{
	register NODEP	np;

	if (opcdp == NNULL) {
		nerr(commap, "Comma list missing opcode!");
	}
	valp = eval(valp);
	if (valp == NNULL) {
		nerr(commap, "comma_list:  rvalue went NULL!");
	}

	valp->n_parent = commap;		/* always true */
	commap->n_right = valp;
	np = opcdp;
	commap->n_tok = np->n_tok;
	commap->n_flags = np->n_flags;
	commap->n_extra.l = np->n_extra.l;

	while (np->n_left != NULL) {		/* find end of list */
		np = np->n_left;
	}

	np->n_left = commap;			/* link */
	commap->n_parent = np;

	return (opcdp);
}


/*
 * c_ofile -- process a file register argument after resolving to a constant
 */

NODEP
c_ofile(register NODEP np)
{
	if (np->n_tok != T_CONST && (np = eval(np))->n_tok != T_CONST) {
		nerr(np, "c_ofile: not a constant!");
	}
	if (np->n_val.l >= FILE_MIN_P0 && np->n_val.l <= FILE_MAX_P0 ||
	    np->n_val.l >= FILE_MIN_P1 && np->n_val.l <= FILE_MAX_P1) {
		np->n_val.l &= FILE_MAX;	/* clip to opcode size */
		return (np);
	}
	nerr(np, "File register 0x%lX is unimplemented on 16C84!",
		np->n_val.l);
	/* NOTREACHED */
	return (NNULL);
}


#ifdef TREE_DEBUG

Static void
tr_pr(NODEP np, int indent)
{
	static char	null_str[] = "";

	if (!np) {
		err("tr_pr:  NULL np!");
		exit(1);
	}
	switch (np->n_tok) {
	case T_CONST:
		printf("%*sC:0x%X\n", indent, null_str, np->n_val.l);
		break;
	case T_B_CONST:
		printf("%*sB:0x%02X, %x\n", indent, null_str,
			np->n_val.l & 0x7F, np->n_val.l >> 7);
		break;
	case T_STRING:
		printf("%*s\"%s\"\n", indent, null_str, np->n_val.str);
		break;
	case T_UNDEF_LABEL:
		printf("%*s%s\n", indent, null_str, np->n_sym->s_name);
		break;
	default:
		printf("%*s%s\n", indent, null_str, tok_name(np->n_tok));
		break;
	}

	indent += 2;
	if (IS_OPER(np)) {
		if (np->n_left) {
			tr_pr(np->n_left, indent);
		}
		if (np->n_right) {
			tr_pr(np->n_right, indent);
		}
	}
	if (IS_CODE(np) && np->n_right) {
		tr_pr(np->n_right, indent);
	}
}


/*
 * tr_pr -- expression tree printer
 */

void
tree_print(NODEP np, char *msg)
{
	printf("%s:\n", msg);
	tr_pr(np, 0);
	(void) fflush(stdout);
}

#endif /* TREE_DEBUG */


/*
 * eval -- recursively evaluate the constant parts of a parse tree
 */

NODEP
eval(NODEP np)
{
	NODEP	lp, rp;

	if (!np)
		return (np);				/* NNULL */

	if (IS_VALUE(np))
		return (np);				/* all done */

	lp = np->n_left;
	rp = np->n_right;

	if (IS_UNIOP(np)) {
		if (rp == NNULL) {
			nerr(np, "eval: unary operator '%s' missing rval!",
				tok_name(np->n_tok));
		}
		if (!IS_CONST(rp)) {
			rp = eval(rp);
#ifdef DEBUG
			if (rp != np->n_right) {
				np->n_right = rp;
				nerr(np, "Right eval changed rp!");
			}
#endif
			if (!IS_CONST(rp))
				return (np);		/* can't eval yet */
		}
	}
	else if (IS_BINOP(np)) {
		if (lp == NNULL) {
			nerr(np, "eval: binary op '%s' missing lval!",
				tok_name(np->n_tok));
		}
		if (rp == NNULL) {
			nerr(np, "eval: binary op '%s' missing rval!",
				tok_name(np->n_tok));
		}
		if (!IS_CONST(lp)) {
			lp = eval(lp);
#ifdef DEBUG
			if (lp != np->n_left) {
				np->n_left = lp;
				nerr(np, "Left eval changed lp!");
			}
#endif
		}
		if (!IS_CONST(rp)) {
			rp = eval(rp);
#ifdef DEBUG
			if (rp != np->n_right) {
				np->n_right = rp;
				nerr(np, "Right eval changed rp!");
			}
#endif
		}
		if (!IS_CONST(lp) || !IS_CONST(rp)) {
			return (np);			/* can't eval yet */
		}
	}
	else if (IS_CODE(np)) {
		if (lp || rp) {
			nerr(np, "eval: no-args opcode has args!");
		}
	}
	else {
		nerr(np, "eval: unknown operator = %d (%s)",
			np->n_tok, tok_name(np->n_tok));
	}

	/*
	 * Arg(s) are constants, so evaluate them.
	 */

	switch (np->n_tok) {
	case T_UMINUS:
		rp->n_val.l = -(rp->n_val.l);
		goto unary_proc;

	case '~':
		rp->n_val.l = ~(rp->n_val.l);
		goto unary_proc;

	case '!':
		rp->n_val.l = !(rp->n_val.l);
		goto unary_proc;

	case '(':				/* paired parenthesis */
		goto unary_proc;

	case T_HIGH:
		rp->n_val.l = (rp->n_val.l >> 8) & 0xFF;
		goto unary_proc;

	case T_LOW:
		rp->n_val.l &= 0xFF;
		goto unary_proc;
unary_proc:
		rp->n_parent = np->n_parent;
		*np = *rp;		/* take attributes from arg */
		free_node(rp);
		return (np);

	case T_MAX:
		lp->n_val.l = MAX(lp->n_val.l, rp->n_val.l);
		goto binary_proc;

	case T_MIN:
		lp->n_val.l = MIN(lp->n_val.l, rp->n_val.l);
		goto binary_proc;

	case '*':
		lp->n_val.l *= rp->n_val.l;
		goto binary_proc;

	case '/':
		if (rp->n_val.l == 0) {
			nerr(rp, "Divide by 0 error! Operator = '/'");
		}
		lp->n_val.l /= rp->n_val.l;
		goto binary_proc;

	case '%':
		if (rp->n_val.l == 0) {
			nerr(rp, "Divide by 0 error! Operator = '%'");
		}
		lp->n_val.l %= rp->n_val.l;
		goto binary_proc;

	case '+':
		lp->n_val.l += rp->n_val.l;
		goto binary_proc;

	case '-':
		lp->n_val.l -= rp->n_val.l;
		goto binary_proc;

	case T_LSHIFT:
		lp->n_val.l <<= rp->n_val.l;
		goto binary_proc;

	case T_RSHIFT:
		lp->n_val.l >>= rp->n_val.l;
		goto binary_proc;

	case '<':
		lp->n_val.l = (lp->n_val.l < rp->n_val.l);
binary_proc:
		lp->n_parent = np->n_parent;
		*np = *lp;		/* take attributes from left arg */
		free_node(lp);
		free_node(rp);
		return (np);

	case '>':
		lp->n_val.l = (lp->n_val.l > rp->n_val.l);
		goto binary_proc;

	case T_LOG_LE:
		lp->n_val.l = (lp->n_val.l <= rp->n_val.l);
		goto binary_proc;

	case T_LOG_GE:
		lp->n_val.l = (lp->n_val.l >= rp->n_val.l);
		goto binary_proc;

	case T_LOG_EQ:
		lp->n_val.l = (lp->n_val.l == rp->n_val.l);
		goto binary_proc;

	case T_LOG_NE:
		lp->n_val.l = (lp->n_val.l != rp->n_val.l);
		goto binary_proc;

	case '&':
		lp->n_val.l &= rp->n_val.l;
		goto binary_proc;

	case '^':
		lp->n_val.l ^= rp->n_val.l;
		goto binary_proc;

	case '|':
		lp->n_val.l |= rp->n_val.l;
		goto binary_proc;

	case T_F_B:
		lp = c_ofile(lp);
		if (rp->n_val.l > 7) {
			nerr(rp, "Invalid bit number = %ld!", rp->n_val.l);
		}
		lp->n_tok = T_B_CONST;			/* bb bfff ffff */
		lp->n_extra.l = BIT_VAL(lp->n_val.l, rp->n_val.l);
		lp->n_val.l = 1 << rp->n_val.l;
		goto binary_proc;

	case T_F_D:
		lp = c_ofile(lp);
		if (rp->n_val.l > 1) {
			nerr(rp, "Invalid destination code = %ld!",
				rp->n_val.l);
		}
		lp->n_val.l = DEST_VAL(lp->n_val.l, rp->n_val.l);
		goto binary_proc;

	/*
	 * Opcodes
	 */
	case T_CALL:		/* these opcodes take an address arg */
	case T_GOTO:
		if (rp->n_val.l > CODE_MAX) {
			nerr(rp, "Invalid code address = 0x%lX!",
				rp->n_val.l);
		}
		goto op_proc;

	case T_ADDLW:		/* these opcodes take a literal arg */
	case T_ANDLW:
	case T_IORLW:
	case T_MOVLW:
	case T_RETLW:
	case T_SUBLW:
	case T_XORLW:
		if (rp->n_val.l > 0xFF) {
			nerr(rp, "Invalid literal value = 0x%lX!",
				rp->n_val.l);
		}
		goto op_proc;

	case T_ADDWF:		/* these opcodes take an f,d arg */
	case T_ANDWF:
	case T_COMF:
	case T_DECF:
	case T_DECFSZ:
	case T_INCF:
	case T_INCFSZ:
	case T_IORWF:
	case T_MOVF:
	case T_RLF:
	case T_RRF:
	case T_SUBWF:
	case T_SWAPF:
	case T_XORWF:
		goto op_proc;

	case T_BCF:		/* these opcodes take an f,b arg */
	case T_BSF:
	case T_BTFSC:
	case T_BTFSS:
		rp->n_val.l = rp->n_extra.l;	/* move f,b value to n_val */
		goto op_proc;

	case T_CLRF:		/* these opcodes take an f arg */
	case T_MOVWF:
		rp = c_ofile(rp);
		goto op_proc;

	case T_CLRW:		/* these opcodes take no arg */
	case T_CLRWDT:
	case T_NOP:
	case T_OPTION:
	case T_RETFIE:
	case T_RETURN:
	case T_SLEEP:
		goto no_arg;

	case T_TRIS:		/* tris is weird */
		rp->n_val.l &= FILE_MAX;
		if (rp->n_val.l < 5 || rp->n_val.l > 6) {
			nerr(rp, "Invalid TRIS register number = %ld!",
				rp->n_val.l);
		}

op_proc:
		if (rp) {
			np->n_opcode |= rp->n_val.l;
		}
no_arg:
		if (code_mem[np->n_addr] >= 0) {
			nerr(np,
				"Code is already defined for addr 0x%03X!"
				" current opcode = 0x%04X, new = 0x%04X",
				np->n_addr,
				code_mem[np->n_addr], np->n_opcode);
		}
		code_mem[np->n_addr] = np->n_opcode;
#ifdef FULL_DEBUG
		printf("coding 0x%04X at 0x%03X\n", np->n_opcode, np->n_addr);
#endif
free_em:
		free_node(rp);
		free_node(np);
		return (NNULL);

	/*
	 * Psuedo-ops
	 */
	case T_DBYTE:
		data_addr += rp->n_val.l;	/* reserve N data bytes */
		rp->n_val.l = data_addr;
		rp = c_ofile(rp);
		goto free_em;

	case T_DWORD:
		data_addr += rp->n_val.l * 2;	/* reserve N data words */
		rp->n_val.l = data_addr;
		rp = c_ofile(rp);
		goto free_em;

	case T_FUSE:
		if (fuse_mem < 0) {
			fuse_mem = rp->n_val.l;
			goto free_em;
		}

		nerr(np, "A fuse byte has already been defined!"
			" current value = 0x%02X, new = 0x%02lX",
			fuse_mem, rp->n_val.l);
		/* NOTREACHED */

	case T_EBYTE:				/* initialize a EE byte */
		if (eedata_mem[np->n_addr] < 0) {
			eedata_mem[np->n_addr] = rp->n_val.l & 0xFF;
			goto free_em;
		}

		nerr(np, "EEPROM data byte 0x%02X has already been defined!"
			" current value = 0x%02X, new = 0x%02lX",
			np->n_addr,
			eedata_mem[np->n_addr], rp->n_val.l);
		/* NOTREACHED */

	case T_EWORD:				/* initialize a EE word */
		if (eedata_mem[np->n_addr] < 0 &&
		    eedata_mem[np->n_addr + 1] < 0) {
			eedata_mem[np->n_addr] = rp->n_val.l & 0xFF;
			eedata_mem[np->n_addr + 1] = (rp->n_val.l >> 8) & 0xFF;
			goto free_em;
		}

		nerr(np, "EEPROM data word 0x%02X has already been defined!"
			" current value = 0x%02X, new = 0x%02lX",
			np->n_addr,
			(eedata_mem[np->n_addr] & 0xFF) |
				(eedata_mem[np->n_addr + 1] & 0xFF << 8),
			rp->n_val.l);
		/* NOTREACHED */

	case T_ID:				/* initialize an ID word */
		if (id_mem[np->n_addr] < 0) {
			id_mem[np->n_addr] = rp->n_val.l & 0x3FFF;
			goto free_em;
		}

		nerr(np, "ID word %d has already been defined!"
			" current value = 0x%02X, new = 0x%02lX",
			np->n_addr, id_mem[np->n_addr], rp->n_val.l);
		/* NOTREACHED */

	case T_BIT:
		if (lp == NNULL) {
			nerr(np, "ERROR: BIT has no label!");
		}
		if (lp->n_tok != T_UNDEF_LABEL) {
			nerr(lp, "ERROR: BIT's label is already defined!");
		}
		if (rp->n_tok != T_B_CONST) {
			nerr(rp, "ERROR: BIT's argument must be a file,bit!");
		}
		label_def(lp, rp->n_extra.l, T_B_CONST);
		free_node(rp);
		free_node(np);
		return (NNULL);

	case T_FILL:				/* fill a block of prog mem */
		if (rp->n_val.l > 0x3FFF) {
			nerr(rp, "Illegal value = 0x%04X in FILL command!",
				rp->n_val.l);
		}
		for ( ; lp->n_val.l != 0; code_addr++, lp->n_val.l--) {
			if (code_addr > CODE_MAX) {
				nerr(np, "Out of PROGRAM memory!");
			}
			if (code_mem[code_addr] >= 0) {
				nerr(lp, "ERROR: (FILL) Code is already "
					"defined for addr 0x%03X!", code_addr);
			}
			code_mem[code_addr] = rp->n_val.l;
#ifdef FULL_DEBUG
			printf("coding 0x%04X at 0x%03X\n",
				rp->n_val.l, np->n_addr);
#endif
		}
		free_node(lp);
		free_node(rp);
		free_node(np);
		return (NNULL);

	case T_EQU:				/* define a permanent label */
		if (lp == NNULL) {
			nerr(np, "ERROR: EQU has no label!");
		}
		if (lp->n_tok != T_UNDEF_LABEL) {
			nerr(lp, "ERROR: EQU's label is already defined!");
		}
		label_def(lp, rp->n_val.l, T_CONST);
		free_node(rp);
		free_node(np);
		return (NNULL);
	}

	nerr(np, "eval: Unknown type = 0x%02X '%s'!",
		np->n_tok, tok_name(np->n_tok));
	/* NOTREACHED */
	return (NNULL);
}


/*
 * label_def -- define a symbol from a label (frees the label NODE)
 */

void
label_def(NODE *label, long val, int type)
{
	register NODE	*np;
	register NODE	*xp;
	SYMBOL		*sym;

	if (label == NULL)
		return;				/* nobody home */

	if (val == -1) {
		val = code_addr;		/* default to code_addr */
	}
	if (label->n_tok != T_UNDEF_LABEL) {
		nerr(label, "Label is already defined!");
	}
	sym = label->n_sym;

	if (!sym) {
		nerr(label, "label_def: has no sym!");
	}
	if (!(sym->s_flags & F_UNDEF_LABEL)) {
		nerr(label, "label_def: node undef, but not symbol!");
	}

	xp = NNULL;
	for (np = sym->s_val.node; np; xp = np, np = np->n_next) {
		if (np == label) {
			if (xp)
				xp->n_next = np->n_next;
			else
				sym->s_val.node = np->n_next;
			break;					/* gotcha */
		}
	}

	if (np == NNULL) {
		nerr(label, "label_def: can't find NODE of '%s' to unlink it!",
			sym->s_name);
	}

	free_node(label);

	np = sym->s_val.node;
	sym->s_val.l = val;
	sym->s_flags &= ~F_UNDEF_LABEL;
	sym->s_type = (uchar) type;

	for ( ; np; np = xp) {
		if (np->n_tok != T_UNDEF_LABEL) {
			nerr(np, "label_def: node not undef!");
		}
		if (np->n_sym != sym) {
			nerr(np, "label_def: not linked to sym!");
		}
		np->n_val.l = val;
		xp = np->n_next;
		np->n_next = NNULL;
		np->n_tok = (uchar) type;
		np->n_flags |= F_IS_CONST;
		/*
		 * Evaluate expression from its root.
		 */
		while (np->n_parent) {
			np = np->n_parent;
		}
		if (!IS_CODE(np)) {
			nerr(np, "resolve: tree not rooted in code! (%s)",
				tok_name(np->n_tok));
#ifdef TREE_DEBUG
			tree_print(np, "Huh?");
#endif
		}
		np = eval(np);
#ifdef TREE_DEBUG
		if (np) {
			tree_print(np, "After resolve");
		}
#endif
	}
}


/*
 * c_f_b -- process file,bit arguments
 *
 * All f,b opcodes are bb bfff ffff, (file | (bit << 7))
 */

NODEP
c_f_b(NODEP filep, NODEP commap, NODEP bitp)
{
	filep->n_parent = bitp->n_parent = commap;
	commap->n_tok = T_F_B;
	commap->n_flags = F_BIN_OP;
	commap->n_left = filep;
	commap->n_right = bitp;
	return (eval(commap));
}


/*
 * c_f_d -- process file,dest arguments (dest is optional, defaults to ",F")
 *
 * All f,d opcodes are dfff ffff, (file | (dest << 7))
 */

NODEP
c_f_d(NODEP filep, NODEP commap, NODEP destp)
{
	filep = eval(filep);
	if (filep == NNULL) {
		if (commap) {
			nerr(commap, "Expected file_reg,dest argument!");
		} else {
			nerr(NNULL, "Expected file_reg,dest argument!");
		}
	}
	if (commap == NNULL) {
		if (filep->n_tok == T_CONST) {
			filep->n_val.l |= 0x80;
			filep->n_flags |= F_IS_CONST;
			return (filep);
		}
		/*
		 * Else, must wait.  Add T_F_D and T_CONST nodes.
		 */
		commap = new_node(T_F_D);
		destp = new_node(T_CONST);
		destp->n_val.l = 1;			/* default = 1 */
		destp->n_flags = F_IS_VAL | F_IS_CONST;
		commap->n_col = destp->n_col = filep->n_col;
	}
	filep->n_parent = destp->n_parent = commap;
	commap->n_tok = T_F_D;
	commap->n_flags = F_BIN_OP;
	commap->n_left = filep;
	commap->n_right = destp;
	return (eval(commap));
}


/*
 * equ_tree_chk -- check a NODE tree for any references to sym
 */
Static int	equ_loop_cnt;

Static int
equ_tree_chk(SYMBOL *sym, NODE *np, NODE *equ)
{
	NODE	*xp;

	if (--equ_loop_cnt < 0) {
		nerr(equ, "Infinite loop resolving EQU!");
	}
	if (np == NNULL)
		return (FALSE);
	if (IS_CONST(np)) {
		if (np->n_tok != T_UNDEF_LABEL)
			return (FALSE);
		if (np->n_sym == sym)
			return (TRUE);
		if (!np->n_sym) {
			nerr(np, "equ_tree_chk: node has no sym!");
		}
		/*
		 * Run down the other symbol's undef list and check for an
		 * T_EQU node that references me.
		 */
		np = np->n_sym->s_val.node;
		while (np) {
			xp = np->n_parent;
			if (!xp) {
				nerr(np, "equ_tree_chk: hit top before EQU!");
			}
			if (xp->n_tok == T_EQU && xp != equ) {
				return (equ_tree_chk(sym, xp->n_right, equ));
			}
			np = np->n_next;
		}

		return (FALSE);
	}
	if (!IS_OPER(np)) {
		nerr(np, "equ_tree_chk: not an operator!");
	}
	if (np->n_left && equ_tree_chk(sym, np->n_left, equ) ||
	    np->n_right && equ_tree_chk(sym, np->n_right, equ)) {
		return (TRUE);
	}
	return (FALSE);
}


/*
 * equ_loop_chk -- check that the given T_EQU node isn't self-referential
 */

Static void
equ_loop_chk(NODE *equ)
{
	register SYMBOL *sym;
	register NODE	*np;

	equ_loop_cnt = 99;
	if (equ->n_tok != T_EQU && equ->n_tok != T_BIT) {
		nerr(equ, "equ_loop_chk: not an EQU or BIT node!");
	}
	np = equ->n_left;
	if (!np) {
		nerr(equ, "equ_loop_chk: NULL n_left!");
	}
	sym = np->n_sym;			/* get label's symbol */
	if (!sym) {
		nerr(np, "equ_loop_chk: NULL sym!");
	}
	if (!(sym->s_flags & F_UNDEF_LABEL)) {
		nerr(np, "equ_loop_chk: sym not undef!");
	}
	for (np = sym->s_val.node; np; np = np->n_next) {
		NODE	*pp;

		if (!(pp = np->n_parent)) {
			nerr(np, "equ_loop_chk: NULL parent!");
		}
		if (pp == equ || (pp->n_tok != T_EQU && pp->n_tok != T_BIT) ||
		    pp->n_left != np)
			continue;
		err("equ_loop_chk: Multiple EQU statements for symbol %s",
			sym->s_name);
	}
	np = equ->n_right;			/* get expression tree */
	if (!np) {
		nerr(equ, "equ_loop_chk: NULL n_right!");
	}
	if (equ_tree_chk(sym, np, equ)) {
		nerr(equ,
			"equ_loop_chk: Self-reference loop in value of '%s'!",
			sym->s_name);
	}
}


/*
 * c_gen_op -- code for a generic opcode (all but weird ones)
 */

NODEP
c_gen_op(NODEP label, NODEP keyw)
{
	long	addr;

	keyw->n_addr = (ushort) (addr = code_addr++);
	if (keyw->n_line && !keyw->n_line->n_extra.sp) {
		keyw->n_line->n_extra.sp = &code_mem[addr];
	}
	(void) eval(keyw);
	if (label) {
		label_def(label, addr, T_CONST);
	}
	return (NNULL);
}


NODEP
c_dorg(NODEP label, NODEP keyw, NODEP arg)
{
	arg = eval(arg);
	if (!arg || arg->n_tok != T_CONST) {
		nerr((arg ? arg : keyw), "DORG requires a constant argument!");
	}
	if (!(arg->n_val.l >= FILE_MIN_P0 && arg->n_val.l <= FILE_MAX_P0 ||
	      arg->n_val.l >= FILE_MIN_P1 && arg->n_val.l <= FILE_MAX_P1)) {
		nerr(arg, "Bad DORG value = 0x%lX (Max = 0x%X)",
			arg->n_val.l, FILE_MAX_P1);
	}
	data_addr = arg->n_val.l;
	if (label) {
		label_def(label, data_addr, T_CONST);
	}
	free_node(keyw);
	free_node(arg);
	return (NNULL);
}


NODEP
c_eorg(NODEP label, NODEP keyw, NODEP arg)
{
	arg = eval(arg);
	if (!arg || arg->n_tok != T_CONST) {
		nerr((arg ? arg : keyw), "EORG requires a constant argument!");
	}
	if (arg->n_val.l >= EEDATA_MAX) {
		nerr(arg, "Bad EORG value = 0x%lX (Max = 0x%X)",
			arg->n_val.l, EEDATA_MAX);
	}
	eedata_addr = arg->n_val.l;
	if (label) {
		label_def(label, eedata_addr, T_CONST);
	}
	free_node(keyw);
	free_node(arg);
	return (NNULL);
}


NODEP
c_org(NODEP label, NODEP keyw, NODEP arg)
{
	arg = eval(arg);
	if (!arg || arg->n_tok != T_CONST) {
		nerr((arg ? arg : keyw), "ORG requires a constant argument!");
	}
	if (arg->n_val.l > CODE_MAX) {
		nerr(arg, "Bad ORG value = 0x%lX (Max = 0x%X)",
			arg->n_val.l, CODE_MAX);
	}
	code_addr = arg->n_val.l;
	if (label) {
		label_def(label, code_addr, T_CONST);
	}
	free_node(keyw);
	free_node(arg);
	return (NNULL);
}


NODEP
c_dbyte(NODEP label, NODEP keyw, NODEP arg)
{
	long	addr;

	arg = eval(arg);
	if (!arg || arg->n_tok != T_CONST) {
		nerr(arg, "DB requires a constant argument!");
	}
	addr = data_addr;
	if (addr > DATA_MAX || (data_addr += arg->n_val.l) > DATA_MAX) {
		nerr(arg, "Out of DATA memory!");
	}
	if (label) {
		label_def(label, addr, T_CONST);
	}
	free_node(keyw);
	free_node(arg);
	return (NNULL);
}


NODEP
c_dword(NODEP label, NODEP keyw, NODEP arg)
{
	long	addr;

	arg = eval(arg);
	if (!arg || arg->n_tok != T_CONST) {
		nerr(arg, "DW requires a constant argument!");
	}
	addr = data_addr;
	if (addr > DATA_MAX || (data_addr += arg->n_val.l * 2) > DATA_MAX) {
		nerr(arg, "Out of DATA memory!");
	}
	if (label) {
		label_def(label, addr, T_CONST);
	}
	free_node(keyw);
	free_node(arg);
	return (NNULL);
}


/*
 * ee_string -- load a string constant into EEDATA, starting at eedata_addr
 */

void
ee_string(NODEP nstr)
{
	register uchar	*up;

	if (nstr->n_line && nstr->n_line->n_extra.sp == NULL) {
		nstr->n_line->n_extra.sp = &eedata_mem[eedata_addr];
	}
	up = (uchar *) nstr->n_val.str;
	while (*up != '\0') {
		if (eedata_addr > EEDATA_MAX) {
			nerr(nstr, "Out of EEPROM memory!");
		}
		if (eedata_mem[eedata_addr] < 0) {
			eedata_mem[eedata_addr] = *up++;
			++eedata_addr;
			continue;
		}

		nerr(nstr, "EEPROM data byte has already been defined!"
			" current value = 0x%02X, new = 0x%02X",
			eedata_mem[eedata_addr], *up);
	}
	free_node(nstr);
}


NODEP
c_ebyte(NODEP label, register NODEP np)
{
	register NODEP	lp;
	register NODEP	rp;
	long		addr = eedata_addr;

	np->n_flags |= F_UNI_OP;

	while (np) {
		np->n_addr = (ushort) eedata_addr;	/* assign addresses */
		if (np->n_addr > EEDATA_MAX) {
			nerr(np, "Out of EEPROM memory!");
		}
		if (np->n_line && np->n_line->n_extra.sp == NULL) {
			np->n_line->n_extra.sp = &eedata_mem[eedata_addr];
		}
		lp = np->n_left;
		np->n_left = NNULL;
		np->n_parent = NNULL;
		if ((rp = np->n_right) && rp->n_tok == T_STRING) {
			ee_string(rp);
			free_node(np);
		} else {
			++eedata_addr;
			(void) eval(np);		/* assign? */
		}
		np = lp;
	}

	if (label) {
		label_def(label, addr, T_CONST);
	}
	return (NNULL);
}


NODEP
c_eword(NODEP label, register NODEP np)
{
	register NODEP	lp;
	register NODEP	rp;
	long		addr;

	addr = eedata_addr;
	np->n_flags |= F_UNI_OP;
	while (np) {
		np->n_addr = (ushort) eedata_addr;	/* assign addresses */
		if (np->n_addr > EEDATA_MAX) {
			nerr(np, "Out of EEPROM memory!");
		}
		if (np->n_line && np->n_line->n_extra.sp == NULL) {
			np->n_line->n_extra.sp = &eedata_mem[eedata_addr];
		}
		lp = np->n_left;
		np->n_left = NNULL;
		np->n_parent = NNULL;
		if ((rp = np->n_right) && rp->n_tok == T_STRING) {
			nerr(rp, "String constant in EEW! (Use EEB instead)");
		}
		eedata_addr += 2;
		(void) eval(np);			/* assign? */
		np = lp;
	}

	if (label) {
		label_def(label, addr, T_CONST);
	}
	return (NNULL);
}


/*
 * id_string -- load a string constant into ID, starting at id_addr
 */

void
id_string(NODEP nstr)
{
	register uchar	*up;

	up = (uchar *) nstr->n_val.str;
	while (*up != '\0') {
		if (id_addr > ID_MAX) {
			nerr(nstr, "Out of ID memory!");
		}
		if (id_mem[id_addr] < 0) {
			id_mem[id_addr] = *up++;
			++id_addr;
			continue;
		}

		nerr(nstr, "ID word has already been defined!"
			" current value = 0x%02X, new = 0x%02X",
			id_mem[id_addr], *up);
	}
	free_node(nstr);
}


/*
 * c_id -- load in ID words
 */

NODEP
c_id(register NODEP np)
{
	register NODEP	lp;
	register NODEP	rp;

	np->n_flags |= F_UNI_OP;

	while (np) {
		np->n_addr = (ushort) id_addr;	/* assign addresses */
		if (np->n_addr > ID_MAX) {
			nerr(np, "Out of ID memory!");
		}
		if (np->n_line && np->n_line->n_extra.sp == NULL) {
			np->n_line->n_extra.sp = &id_mem[id_addr];
		}
		lp = np->n_left;
		np->n_left = NNULL;
		np->n_parent = NNULL;
		if ((rp = np->n_right) && rp->n_tok == T_STRING) {
			id_string(rp);
			free_node(np);
		} else {
			++id_addr;
			(void) eval(np);		/* assign? */
		}
		np = lp;
	}

	return (NNULL);
}


/*
 * retlw_string -- load a string constant into CODE mem, starting at code_addr
 */

void
retlw_string(NODEP nstr)
{
	register uchar	*up;

	up = (uchar *) nstr->n_val.str;
	while (*up != '\0') {
		if (code_addr > CODE_MAX) {
			nerr(nstr, "Out of PROGRAM memory!");
		}
		if (code_mem[code_addr] < 0) {
			code_mem[code_addr] = OP_RETLW | *up++;
#ifdef FULL_DEBUG
			printf("coding 0x%04X at 0x%03X\n",
				code_mem[code_addr], code_addr);
#endif
			++code_addr;
			continue;
		}

		nerr(nstr, "Code is already defined for addr 0x%03X!"
			" current opcode = 0x%04X",
			code_addr, code_mem[code_addr]);
	}
	free_node(nstr);
}


/*
 * c_retlw -- turn a RETLW opcode's argument(s) into individual nodes
 */

NODEP
c_retlw(NODEP label, register NODEP np)
{
	register NODEP	lp;
	register NODEP	rp;
	long		addr;

	np->n_flags |= F_UNI_OP;
	addr = code_addr;

	while (np) {
		np->n_addr = (ushort) code_addr;	/* assign addresses */
		if (np->n_addr > CODE_MAX) {
			nerr(np, "Out of PROGRAM memory!");
		}
		if (np->n_line && np->n_line->n_extra.sp == NULL) {
			np->n_line->n_extra.sp = &code_mem[code_addr];
		}
		lp = np->n_left;
		np->n_left = NNULL;
		np->n_parent = NNULL;
		if ((rp = np->n_right) && rp->n_tok == T_STRING) {
			retlw_string(rp);
			free_node(np);
		} else {
			++code_addr;
			(void) eval(np);		/* assign? */
		}
		np = lp;
	}

	if (label) {
		label_def(label, addr, T_CONST);
	}

	return (NNULL);
}


NODEP
c_bit(NODEP label, NODEP keyw, NODEP arg)
{
	if (!label) {
		nerr(keyw, "Must give BIT a symbol!");
	}
	if (label->n_tok != T_UNDEF_LABEL) {
		nerr(label, "Tried to BIT an already defined symbol!");
	}
	keyw->n_left = label;
	keyw->n_right = arg;
	keyw->n_flags |= F_BIN_OP;
	arg->n_parent = label->n_parent = keyw;
	label->n_flags |= F_IS_CONST;	/* This is a necessary lie. */
	keyw = eval(keyw);
	if (keyw)
		equ_loop_chk(keyw);
	return (NNULL);
}


NODEP
c_equ(NODEP label, NODEP keyw, NODEP arg)
{
	if (!label) {
		nerr(keyw, "Must give EQU a symbol!");
	}
	if (label->n_tok != T_UNDEF_LABEL) {
		nerr(label, "Tried to EQU an already defined symbol!");
	}
	keyw->n_tok = T_EQU;			/* fix '=' variant */
	keyw->n_left = label;
	keyw->n_right = arg;
	keyw->n_flags |= F_BIN_OP;
	label->n_parent = arg->n_parent = keyw;
	label->n_flags |= F_IS_CONST;	/* This is a necessary lie. */
	keyw = eval(keyw);
	if (keyw)
		equ_loop_chk(keyw);
	return (NNULL);
}


/*
 * c_unequ -- undefine a symbol
 */

NODEP
c_unequ(NODEP keyw, NODEP arg)
{
	register SYMBOL *targ;
	register SYMBOL *sp;
	register SYMBOL **hashp;
	register SYMBOL *prev;

	/*
	 * Defined symbols will have a n_tok of T_CONST and a n_extra that
	 * points to its symbol.  Ordinary constants will have a NULL
	 * n_extra.  Undefined symbols will still be UNDEF.
	 */
	targ = arg->n_sym;
	switch (arg->n_tok) {
	case T_CONST:
		if (!targ) {
			nerr(arg, "Can't UNEQU a numeric constant!");
		}
		if (targ->s_flags & F_UNDEF_LABEL) {
			nerr(arg, "Can't UNEQU an undefined symbol!");
		}
		if (targ->s_val.l != arg->n_val.l) {
			nerr(arg, "Symbol and node values don't match!");
		}
		break;
	case T_UNDEF_LABEL:
		nerr(arg, "Can't UNEQU an undefined or partially defined "
			"symbol!");
		/* NOTREACHED */
	default:
		nerr(arg, "Can't UNEQU a non-symbol!  (tok = 0x%02X, '%s')",
			arg->n_tok, tok_name(arg->n_tok));
	}

	hashp = SYM_HASH(targ->s_name);
	for (prev = NULL, sp = *hashp; sp; prev = sp, sp = sp->s_next) {
		if (sp == targ) {
			break;
		}
	}

	if (sp == NULL) {
		nerr(arg, "UNEQU: didn't find symbol '%s'!", sp->s_name);
	}

	if (prev == NULL) {
		*hashp = targ->s_next;
	}
	else {
		prev->s_next = targ->s_next;
	}

	free(targ);
	free_node(arg);
	return (NNULL);
}


/*
 * c_fill -- fill num words of program memory with the given value
 */

NODEP
c_fill(NODEP label, NODEP keyw, NODEP nump, NODEP commap, NODEP valp)
{
	long	addr;

	free_node(commap);
	addr = keyw->n_addr = code_addr;
	keyw->n_flags |= F_BIN_OP;
	nump->n_parent = valp->n_parent = keyw;
	keyw->n_left = nump;
	keyw->n_right = valp;
	(void) eval(keyw);
	if (label) {
		label_def(label, addr, T_CONST);
	}
	return (NNULL);
}


/*
 * c_fuse -- set the configuration fuse word
 */

NODEP
c_fuse(NODEP keyw, NODEP valp)
{
	valp = eval(valp);
	if (valp->n_tok != T_CONST) {
		nerr(valp, "FUSE requires a constant argument!");
	}
	if (fuse_mem >= 0) {
		nerr(valp, "The FUSE word is already programmed! (0x%02lX)",
			(ulong) fuse_mem);
	}
	fuse_mem = (short) valp->n_val.l;
	if (keyw->n_line && keyw->n_line->n_extra.sp == NULL) {
		keyw->n_line->n_extra.sp = &fuse_mem;
	}

	free_node(keyw);
	free_node(valp);
	return (NNULL);
}


/*
 * c_dol -- set the value of the '$' token
 */

NODEP
c_dol(NODEP np)
{
	np->n_val.l = code_addr;		/* current prog addr */
	np->n_tok = T_CONST;
	np->n_flags |= F_IS_VAL | F_IS_CONST;
	return (np);				/* all done */
}


/*
 * lex interface
 */

int
yywrap(void)
{
	return (1);		/* let flex return EOF */
}


/*
 * sym_cmp_name -- qsort comparison function
 */

Static int
sym_cmp_name(const void *a, const void *b)
{
	return (cmp_ncase((*(SYMBOL **)a)->s_name, (*(SYMBOL **)b)->s_name));
}


/*
 * sym_cmp_val -- qsort comparison function
 */

Static int
sym_cmp_val(const void *ap, const void *bp)
{
	SYMBOL	*a = *(SYMBOL **)ap;
	SYMBOL	*b = *(SYMBOL **)bp;
	int	val;

	val = (int)(a->s_val.l - b->s_val.l);
	return (val ? val : strcmp(a->s_name, b->s_name));
}


/*
 * sym_output -- dump the symbol table, while checking for undefined symbols
 */

Static int
sym_output(void)
{
	register SYMBOL	*sp;
	register SYMBOL	*xp;
	register NODE	*np;
	register SYMBOL	**ap;
	register int	cnt;
	register int	i;
	register int	n;
	SYMBOL		*head;
	int		error;
	FILE		*fp;
	char		str[128];

	fp = fopen(lst_file, "w");
	if (!fp) {
		err("Can't open listing file!");
		exit(1);
	}

	n = 0;
	for (np = list_head; np; np = np->n_next) {
		register short	*shp;

		++n;
		if (shp = np->n_extra.sp) {
			if (shp < eedata_mem) {
				fprintf(fp, "%4d %03X:%04X\t%s\n",
					n, (shp - code_mem), *shp,
					np->n_val.str);
			} else if (shp < id_mem) {
				fprintf(fp, "%4d E%02X:%04X\t%s\n",
					n, (shp - eedata_mem), *shp,
					np->n_val.str);
			} else if (shp < &fuse_mem) {
				fprintf(fp, "%4d ID%X:%04X\t%s\n",
					n, (shp - id_mem), *shp,
					np->n_val.str);
			} else {
				fprintf(fp, "%4d FUS:%04X\t%s\n",
					n, *shp, np->n_val.str);
			}
		}
		else {
			fprintf(fp, "%4d\t\t%s\n", n, np->n_val.str);
		}
	}
	fprintf(fp, "\n");

	error = 0;
	cnt = 0;
	head = (SYMBOL *) NULL;
	ap = &sym_hash[0];
	for (n = NUM_HASH; --n >= 0; ap++) {
		for (sp = *ap; sp; sp = xp) {
			xp = sp->s_next;
			if (sp->s_flags & F_DONT_LIST) {
				continue;			/* no show */
			}
			if (!(sp->s_flags & F_UNDEF_LABEL)) {
				++cnt;
				sp->s_next = head;		/* link */
				head = sp;
				continue;			/* more! */
			}
			/*
			 * Complain about undefined symbols.
			 */
			++error;
			np = sp->s_val.node;
			if (np == NNULL) {
				errno = 0;
				err("Huh? UNDEF symbol '%s' has no nodes!",
					sp->s_name);
				continue;
			}
			(void) sprintf(str,
				"Undefined symbol: %s "
				"(first used line # %ld, char # %d)\n",
				sp->s_name, sp->s_lineno, sp->s_col);
			errno = 0;
			err("%s", str);
			(void) fprintf(fp, str);
#ifdef FULL_DEBUG
			for (i = 0; np; i++, np = np->n_next)
				;
			(void) fprintf(fp, "\tHas %d NODEs\n", i);
#endif
		}
	}
	if (cnt == 0) {
		return (error);
	}

	ap = (SYMBOL **) alloc(cnt * sizeof(sp), "symbol sort table");

	cnt = 0;
	for (sp = head; sp; sp = sp->s_next) {
		ap[cnt++] = sp;				/* put in array */
	}

	qsort((void *)ap, cnt, sizeof(sp), sym_cmp_name);

	(void) fprintf(fp, "\n Symbol Table (sorted by name)\n");
	for (i = 0; i < cnt; i++) {
		sp = ap[i];
		(void) fprintf(fp, "%08lX= %s\n", sp->s_val.l, sp->s_name);
	}

	qsort((void *)ap, cnt, sizeof(sp), sym_cmp_val);

	(void) fprintf(fp, "\n Symbol Table (sorted by value)\n");
	for (i = 0; i < cnt; i++) {
		sp = ap[i];
		(void) fprintf(fp, "%08lX@ %s\n", sp->s_val.l, sp->s_name);
	}

	(void) fclose(fp);
	return (error);
}


/*
 * output_mem -- output a memory space
 */

Static void
output_mem(register short *shp, register int addr, int num)
{
	register int	run;

	run = 0;
	while (--num >= 0) {
		if (*shp < 0) {
			++addr;
			++shp;
			run = 0;
			continue;
		}
		if (run == 0) {
			++run;
			/* The "2 * addr" is an artifact of INHX8M format */
			i8_addr(2 * addr);
		}
		++addr;
		i8_word(*shp++);
	}
}


Static void
output(void)
{
	if (sym_output())
		return;				/* undef syms -- no output */

	i8_open(hex_file);

	output_mem(code_mem, CODE_START, NEL(code_mem));

	output_mem(id_mem, ID_START, NEL(id_mem));

	if (fuse_mem >= 0) {
		output_mem(&fuse_mem, FUSE_ADDR, 1);
	}

	output_mem(eedata_mem, EEDATA_START, NEL(eedata_mem));

	i8_close();
}


/*
 * init_symbols -- initialize the symbol table with 16C84 registers, etc.
 */

Static void
init_symbols(void)
{
	register SYMBOL		*sp;
	register SYMBOL		**hashp;
	register SYM_INIT	*sip;
	register int		n;
	register int		len;

	for (sip = sym_init, n = NEL(sym_init); --n >= 0; sip++) {
		len = strlen(sip->si_name);
		sp = (SYMBOL *) alloc(sizeof(SYMBOL) + len, "symbol init");
		bzero(sp, sizeof(*sp));
		(void) strncpy(sp->s_name, sip->si_name, len);
		sp->s_name[len] = '\0';
		hashp = SYM_HASH(sp->s_name);
		sp->s_next = *hashp;			/* link on hash list */
		*hashp = sp;
		sp->s_flags = F_DONT_LIST;
		sp->s_val.l = sip->si_value;
		sp->s_type = sip->si_tok;
	}
}


/*
 * init -- init stuff
 */

Static void
init(int argc, char **argv)
{
	register char	*p;
	extern FILE	*yyin;

	printf("\nasm_c84 -- PIC 16C84 assembler.\n\n"
		"Copyright (C) James Cleverdon.  Freely distributable.\n"
		"(Covered by the \"GNU Copyleft\")\n\n");

	errno = 0;
	if (--argc <= 0) {
		err("Usage:  %s PIC_16C84_asm_file", argv[0]);
		err("\n\tAssembler for the Microchip Technologies PIC 16C84");
		err("\tEEPROM microcontroller.");
		exit(1);
	}

	/*
	 * Open input file.
	 */
	++argv;
	this_file = p = argv[0];
	yyin = fopen(p, "r");
	if (yyin == NULL) {
		err("Can't open 16C84 assembly file '%s'!", p);
		exit(1);
	}

	/*
	 * Make XXX.hex and XXX.lst file names.
	 */
	if (p = strrchr(p, '.')) {
		argc = (p - this_file) + 4 + 1;
	} else {
		argc = strlen(p) + 4 + 1;
	}
	lst_file = alloc(argc, "lst_file string");
	hex_file = p = alloc(argc, "hex_file string");
	(void) strcpy(p, this_file);
	if (p = strrchr(p, '.')) {
		*p = '\0';			/* zap .???? suffix */
	}
	(void) sprintf(lst_file, "%s.lst", hex_file);
	(void) sprintf(hex_file, "%s.hex", hex_file);

	memset((void *)code_mem, -1, sizeof(code_mem));
	memset((void *)eedata_mem, -1, sizeof(eedata_mem));
	memset((void *)id_mem, -1, sizeof(id_mem));
	fuse_mem = -1;

	init_symbols();
}


/*
 * main -- main routine
 */

void
main(int argc, char **argv)
{
	init(argc, argv);
	(void) yyparse();
	output();
	exit(0);
}

extern int		sys_nerr;


/*
 * nerr -- print out a verbose message, then exit
 */

void
nerr(NODE *np, char *fmt, ...)
{
	NODE		*lp;
	register char	*p;
	va_list		vap;

	(void) fflush(stdout);

	if (np == NNULL) {
		np = new_node(0);			/* fake it */
		np->n_lineno = line_num;
		np->n_col = col;
		np->n_file = this_file;
		np->n_line = list_tail;
	}

	if (lp = np->n_line) {
		fprintf(stderr, "%s\n", lp->n_val.str);
		/*
		 * To avoid TAB vs. space column differences, convert all
		 * non-whitespace chars between 0 and n_col into spaces.
		 */
		p = &lp->n_val.str[np->n_col];
		p[0] = '^';
		p[1] = '\0';
		while (--p >= lp->n_val.str) {
			if (!isspace(*p))
				*p = ' ';
		}
		fprintf(stderr, "%s\n", lp->n_val.str);
	}

	fprintf(stderr, "file '%s', line # %ld, char # %d\n\n",
		np->n_file, np->n_lineno, np->n_col);

	va_start(vap, fmt);
	(void) vfprintf(stderr, fmt, vap);
	va_end(vap);
	(void) fprintf(stderr, "\n");

	(void) fflush(stderr);
	exit(1);
	/* NOTREACHED */
}


/*
 * lex and yacc's error message print function
 */

void
yyerror(char *s)
{
	nerr(NNULL, "yyerror:  %s", s);
}


/*
 * err -- print error message
 */

void
err(char *fmt, ...)
{
	int	errnum = errno;
	va_list	vap;

	(void) fflush(stdout);
	va_start(vap, fmt);
	(void) vfprintf(stderr, fmt, vap);
	va_end(vap);
	if (errnum < 0 || errnum > sys_nerr)
		(void) fprintf(stderr, " (err=%d)\n", errnum);
	else if (errnum)
		(void) fprintf(stderr, " (err=%d: %s)\n",
				errnum, strerror(errnum));
	else {
		(void) fprintf(stderr, "\n");
	}
	(void) fflush(stderr);
}
