
/*
 * emitter.c -- emit INtel HeX 8-bit Merged (INHX8M) records.
 *
 * Copyright (C) 1995-6 James Cleverdon
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Address:
 *	James Cleverdon
 *	16465 S.W. Estuary Dr. # 101
 *	Beaverton, OR 97006-7954, USA
 *
 *	jamesc@sequent.com
 *
 * $Log:	emitter.c,v $
 * Revision 1.0  95/12/19  01:28:34  jamesc
 * "CSE510_project"
 * 
 */

#ident "$Id: emitter.c,v 1.0 95/12/19 01:28:34 jamesc Exp Locker: jamesc $"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "c84_ext.h"

/*
 * Intel hex file format.
 *
 * Record format (all fields are hexadecimal):
 *
 * :NNAAAATTXXXX...XXXCC\n
 *
 * Where:
 *	:	-- start character
 *	NN	-- byte count (some burners only accept limited lengths)
 *	AAAA	-- hex byte address (hi/low, i.e. 0x1234 appears as "1234")
 *	TT	-- record type code (00 = data, 01 = end-of-data)
 *	XX	-- data byte(s)
 *	XXXX	-- data words (low/hi, i.e. 0x1234 appears as "3412")
 *	CC	-- one-byte 1s complement checksum
 *	\n	-- newline character (CTRL-J, or whatever)
 */

#define I_REC_NUM	16			/* maximum bytes per I-rec */
#define I_OFFSET_MAX	0xFFFF			/* maximum 2-byte address */

static ulong	addr = 0;
static uint	pos = 0;
static FILE	*fout = NULL;
static uchar	obuf[I_REC_NUM + 4];
static char	err_str[] = "I-rec fprintf failed!";

extern int	errno;

extern void	err(char *fmt, ...);


/*
 * i8_open -- open an Intel Hex output file and prepare to write it
 */

void
i8_open(char *file)
{
	errno = 0;
	if (fout != NULL) {
		err("i8_open:  Called while busy!");
		exit(1);
	}

	if ((fout = fopen(file, "w")) == NULL) {
		err("Cannot open output file '%s' for writing! (err # %d)",
			file, errno);
		exit(1);
	}

	pos = 0;
	addr = 0;
}


static uint
hexbytes(register uchar *up, register int len, register uint chksum)
{
	while (--len >= 0) {
		chksum += *up;
		if (fprintf(fout, "%02X", *up++) <= 0) {
			err(err_str);
			exit(1);
		}
	}
	return (chksum);
}


static void
outline(void)
{
	int	chksum;
	uchar	buf4[4];

	if (fprintf(fout, ":") <= 0) {			/* start char */
		err(err_str);
		exit(1);
	}

	buf4[0] = pos;					/* byte count */
	buf4[1] = (addr >> 8) & 0xFF;			/* address (hi/low) */
	buf4[2] = addr & 0xFF;
	buf4[3] = 0;					/* type:  data rec */

	chksum = hexbytes(buf4, 4, 0);
	chksum = hexbytes(obuf, pos, chksum);

							/* 1s comp checksum */
	if (fprintf(fout, "%02X\n", (-chksum) & 0xff) <= 0) {
		err(err_str);
		exit(1);
	}

	addr += pos;
	pos = 0;
}


/*
 * i8_close -- flush any output, add an end-of-data line, and close the file
 */

void
i8_close(void)
{
	if (pos)
		outline();

	if (fprintf(fout, ":00000001FF\n") <= 0) {	/* end of data rec */
		err(err_str);
		exit(1);
	}
	fclose(fout);
	fout = NULL;
}


/*
 * i8_addr -- Called to set a new address in the assembler (i.e. .org, .skip)
 *
 *	Flushes any current output first.
 */

void
i8_addr(ulong a)
{
	if (pos)
		outline();

	addr = a & I_OFFSET_MAX;
}


/*
 * i8_byte -- called with each byte to be output
 */

void
i8_byte(int b)
{
	if (pos >= I_REC_NUM)
		outline();

	obuf[pos++] = b & 0xFF;
}


/*
 * i8_word -- do a 16-bit INHX8M word as two 8-bit bytes (low/hi)
 */

void
i8_word(int w)
{
	if (pos >= I_REC_NUM - 1)
		outline();

	obuf[pos++] = w & 0xFF;
	obuf[pos++] = (w >> 8) & 0xFF;
}
